
#ifndef _ARRAY_H_
#define _ARRAY_H_


/* Bear Transfer Protocol (BTP), a TCP/IP socket based file 
   transfer utility

  The author of this software is Ian Kaplan
  www.bearcave.com
  iank@bearcave.com

  Copyright (c) Ian Kaplan, 1999

  See "copyright" file for allowed use.

*/

/*

   This file defines an array template class that will grow as
   elements are added to the end of the array.

   This array class is designed for dense arrays where the all
   elements of the array are used.

   Usage:

     - Elements are added to the end of the array via the append
       function.
     - Elements in the array can be accessed via the [] operator.
     - If the elements in the array are dynamicly allocated, the
       user is responsible for deallocating these elements.
     - The elements of the array cannot be a class with virtual
       functions.

 */

template <class T>
class GrowableArray {
private:
  typedef enum { StartArraySize = 128 } bogus;
  void *(*pGetMemFx)(const unsigned int size);
  void (*pFreeMemFx)( void *);
  size_t num_elem;
  size_t array_size;
  T *pArray;

public:
  GrowableArray( void *(*GetMem)(const unsigned int size ),
		 void (*FreeMem)( void *) )
  {
    pGetMemFx = GetMem;
    pFreeMemFx = FreeMem;

    pArray = (T *)(*pGetMemFx)( sizeof( T ) * StartArraySize );
    num_elem = 0;
    array_size = StartArraySize;
  } // GrowableArray constructor

  ~GrowableArray()
  {
    (*pFreeMemFx)( (void *)pArray );
  } // GrowableArray destructor

  size_t get_size(void) { return num_elem; }

  // LHS [] operator
  T &operator[](const size_t i)
  {
    assert( i < num_elem );
    return pArray[ i ];
  }

  // RHS [] operator
  T operator[](const size_t i ) const
  {
    assert( i < num_elem );
    return pArray[ i ];
  }

  // append an item to the end of the array
  void append( T item )
  {

    if (num_elem == array_size) {
      T *old_array;
      size_t new_size, i;

      old_array = pArray;
      new_size = array_size * 2;

      // allocate a new array that is twice the size of the old one
      pArray = pArray = (T *)(*pGetMemFx)( sizeof( T ) * new_size );

      for (i = 0; i < array_size; i++) {
	pArray[i] = old_array[i];
      }

      // free the old array
      (*pFreeMemFx)( old_array );
      // new size is old size times two
      array_size = new_size;
    }

    pArray[ num_elem ] = item;
    num_elem++;
  } // append

}; // GrowableArray

#endif
