
/* Bear Transfer Protocol (BTP), a TCP/IP socket based file 
   transfer utility

  The author of this software is Ian Kaplan
  www.bearcave.com
  iank@bearcave.com

  Copyright (c) Ian Kaplan, 1999

  See "copyright" file for allowed use.

*/

/*
 * commands.C
 *

   Class functions to process the BTP command prompt input

 */

#include "netinc.h"

#include "neterr.h"
#include "netwrap.h"
#include "iowrap.h"
#include "commands.h"
#include "netio.h"



/*
 * skip_leading_white_spaces
 *
 */
void client_commands::skip_leading_white_spaces(void)
{
  char ch;
  while ( (ch = get_char()) != '\0' && isspace(ch))
    /* nada */;

  if (ch != '\0')
    unget_char();
} // skip_leading_white_spaces


/*
 * is_yes
 *

   The function returns TRUE if the user enters 'y', otherwise FALSE.
   This function is used to support the y/[n] prompt (e.g., y or
   default n).

 */
Boolean client_commands::is_yes()
{
  Boolean rslt;
  char AnsChr;

  AnsChr = getchar();

  // skip the \n so that the next read will not pick it up
  if (AnsChr != '\n') {
    while (getchar() != '\n' && (! feof( stdin )))
      /* nada */;
  }
  
  rslt = (Boolean)(AnsChr == 'y');

  return rslt;
}  // is_yes



/*
 * Fgets
 */
char *client_commands::Fgets( char *buf, const int bufsize, FILE *fp )
{
  char *pStr = NULL;

  init();
  if ((pStr = fgets( buf, bufsize, fp)) != NULL) {
    init_buflen();
    init_charptr();
  }
  return pStr;
}


/*
 * get_line
 *
 */
void client_commands::get_line(void)
{
  int done;

  done = FALSE;

  while (! done ) {
    netio io;
    COMMANDS mode;

    printf("btp (");
    io.print_mode();
    printf(")=> "); // btp: bear transfer protocol
    if (Fgets( linebuf, sizeof(linebuf), stdin) != NULL) {
      skip_leading_white_spaces();
      if (get_buflen() > 0)
	done = TRUE;
    }
    else {
      // in theory control-D should shut down the
      // stream on UNIX, resulting in a NULL line
      line_error = TRUE;
      done = TRUE;
    }
  }
  
} // get_line



/*
 * get_word
 *
 */
int client_commands::get_word( char *pWordBuf, const int word_size )
{
  int rslt;

  rslt = FALSE;

  if (pWordBuf != NULL) {
    int ix;
    char ch;

    ix = 0;
    skip_leading_white_spaces();

    if (get_buflen() > 0) {
      while ((ch = get_char()) != '\0' && (! isspace(ch)) && ix < word_size) {
	pWordBuf[ix] = ch;
	ix++;
      }

    }
    if (ix < word_size)
      pWordBuf[ix] = '\0';
    else
      pWordBuf[word_size-1] = '\0';

    if (strlen( pWordBuf ) > 0)
      rslt = TRUE;
  }
  return rslt;
} // get_word



/*
 * find_command
 *

   This function is passed a word from the BTP prompt line.
   It searches the command table for a matching command.
   If a command is found it returns the associated enumeration.

   To add a new command, add the enumeration to commands.h
   and add the word and associated enumeration to the table
   below.  For debug add the command to the comand_name table
   function (below) as well.

 */
COMMANDS client_commands::find_command( const char *pWord )
{
  typedef struct {
    char *command_name;
    COMMANDS com;
  } table_ent;

  static table_ent command_table[] = { {"quit",   quit},
				       {"get",    get_file },
				       {"put",    put_file },
				       {"pwd",    print_cwd },
				       {"cd",     change_dir },
				       {"ls",     list_dir },
				       {"local",  local_command },
				       {"ascii",  ascii_mode },
				       {"binary", binary_mode },
                                       {"help",   help },
				       {"about",  about },
				       {"debug",  debug_mode },
				       {"on",     turn_on },
				       {"off",    turn_off },
				       { NULL,    bad_command }
				     };

  COMMANDS com;

  com = bad_command;
  if (pWord != NULL) {
    table_ent *pTableEnt;

    pTableEnt = command_table;

    while (pTableEnt->command_name != NULL) {
      if (strcmp( pTableEnt->command_name, pWord) == 0) {
	com = pTableEnt->com;
	break;
      }
      pTableEnt++;
    } // while
  } // if

  return com;
} // find_command



/*
 * get_command
 *
 */
COMMANDS client_commands::get_command()
{
  COMMANDS com;
  char word[ 80 ];

  com = quit;
  if (! line_error) {
    if (get_word( word, sizeof( word ) )) {
      com = find_command( word );
    }
  }
    
  return com;
} // get_command



/*
 * command_name
 *

   Return a string for the name of a command enumeration

 */
const char *client_commands::command_name( COMMANDS com )
{
  static char *names[] = { "bad command",
			   "put file",
			   "get file",
			   "acknowledge",
			   "quit",
			   "send name",
			   "file exists",
			   "directory",
			   "overwrite file",
			   "no_such_file",
			   "com_error",
			   "could_not_open",
			   "sending_data",
			   "send_done",
			   "print_cwd",
			   "change_dir",
			   "local",
			   "list_dir",
			   "ascii mode",
			   "binary mode",
			   "help",
			   "about",
			   "debug",
			   "debug on",
			   "debug off"
			   "on",
			   "off",
			   NULL };

  const char *retstr = "unknown command";

  if (com > bad_command && com < unknown_command) {
    retstr = names[ com ];
  }

  return retstr;
} // client_commands
