
/* 
   Bear Transfer Protocol (BTP), a TCP/IP socket based file 
   transfer utility

  The author of this software is Ian Kaplan
  www.bearcave.com
  iank@bearcave.com

  Copyright (c) Ian Kaplan, 1999

  See "copyright" file for allowed use.

*/
/*

   BTP Client

 */

/*

   To compile use Makefile

 */

#include "netinc.h"
#include <sys/timeb.h>

#include "neterr.h"
#include "netwrap.h"
#include "iowrap.h"
#include "commands.h"
#include "netio.h"
#include "interval.h"
#include "netfile.h"
#include "switch.h"
#include "netserv.h"
#include "heapsort.h"
#include "array.h"
#include "dirnames.h"
#include "list.h"
#include "strheap.h"
#include "readdir.h"
#include "comline.h"
#include "ftp_client.h"



/*
 * ok_to_overwrite
 *

   Ask user if its all right to overwrite a file.

 */
Boolean ftp_client::ok_to_overwrite( const char *file_name )
{
  printf("overwrite %s? (y/[n]) ", file_name );

  return com.is_yes();
} // ok_to_overwrite



/*
 * open_server_dest
 *

   Open "destname" file on server.  If the server file is opened,
   return TRUE, else return FALSE.

   The algorithm followed is:

     - If the destination name does not exist on the remote system
       return TRUE.

     - If the destination name does exist, and its not a directory,
       ask the user if its OK to overwrite the file.  If the user says
       yes, return TRUE, otherwise return FALSE.

 */
Boolean ftp_client::open_server_dest( int sockfd, const char *destname, 
				      const char *prog_name )
{
  Boolean ok;
  COMMANDS com_in;
  int len;

  assert( destname != NULL );

  // send put_file command to remote system
  netio_obj.SendCommand( sockfd, put_file );
  netio_obj.GetAck( sockfd, put_file, prog_name );  // get ack.

  // send send_name (e.g., client will be sending name) command
  // to remote system
  netio_obj.SendCommand( sockfd, send_name );
  netio_obj.GetAck( sockfd, send_name, prog_name );
  
  // Send the destination file name to the remote system
  len = strlen( destname ) + 1; // pick up the null terminator
  netio_obj.Writen( sockfd, (void*)destname, len );

  // Remote system will see if the destname file exists
  ok = TRUE;
  com_in = netio_obj.RecvCommand( sockfd, prog_name);

  if (com_in != acknowledge) {
    ok = FALSE;
    switch (com_in) {
    case directory:
      printf("%s: %s is a directory\n", prog_name, destname );
      break;
    case file_exists:
      // get user answer
      if (ok_to_overwrite( destname )) {
	netio_obj.SendCommand( sockfd, overwrite_file );
	com_in = netio_obj.RecvCommand( sockfd, prog_name);
	if (com_in != acknowledge) {
	  printf("%s: expecting ack. from overwrite_file command\n",
		 prog_name );
	}
	else
	  ok = TRUE;
      }
      else {
	// we don't want to overwrite the destination file, so
	// tell the server to continue
	netio_obj.SendCommand( sockfd, acknowledge );
      }
      break;

    case could_not_open:
      printf("%s: could not open file %s\n", prog_name, destname );
      break;

    case com_error:
      printf("%s: put_transfer - communication error\n", prog_name );
      break;

    default:
      printf("%s: put_transfer - unknown return status\n", prog_name );
      break;
    } // switch
  }

  return ok;
}  // open_server_dest




/*
 * open_server_src
 *

   Attempt to open the source file on the server system.  If the
   file could be opened, the function returns TRUE, otherwise
   FALSE.

 */
Boolean ftp_client::open_server_src( int sockfd, const char *srcname, 
				     const char *prog_name)
{
  Boolean ok;
  COMMANDS com_in;
  int len;

  // send get_file command to remote system
  netio_obj.SendCommand( sockfd, get_file );
  netio_obj.GetAck( sockfd, get_file, prog_name );  // get ack.

  // send send_name (e.g., client will be sending name) command
  // to remote system
  netio_obj.SendCommand( sockfd, send_name );
  netio_obj.GetAck( sockfd, send_name, prog_name );

  // Send the destination file name to the remote system
  len = strlen( srcname ) + 1; // pick up the null terminator
  netio_obj.Writen( sockfd, (void*)srcname, len);

  // Remote system will see if the destname file exists
  ok = TRUE;
  com_in = netio_obj.RecvCommand( sockfd, prog_name);

  if (com_in != acknowledge) {
    ok = FALSE;
    switch (com_in) {
    case directory:
      printf("%s: %s is a directory on server system\n", prog_name, srcname );
      break;

    case no_such_file:
      printf("%s: file %s not found on server\n", prog_name, srcname );
      break;

    case could_not_open:
      printf("%s: could not open file %s on server\n", prog_name, srcname );
      break;

    case com_error:
      printf("%s: put_transfer - communication error\n", prog_name );
      break;

    default:
      printf("%s: put_transfer - unknown return status\n", prog_name );
      break;
    } // switch
  }

  return ok;
} // open_server_src


/*
 * server_cd
 *

   This function is called to process the "cd" command.  The "cd" part
   of the command has been parsed before the call.  If there is a
   "word" following the "cd" giving the path, the path is sent
   to the server, which will attempt to make the path the current
   working directory.

 */
void ftp_client::server_cd( int sockfd, const char *prog_name )
{
  char new_wd[ 128 ];

  if (com.get_word( new_wd, sizeof( new_wd ))) {
    COMMANDS cmd;
    int len;

    netio_obj.SendCommand( sockfd, change_dir );
    netio_obj.GetAck( sockfd, change_dir, prog_name );  // get ack.

    len = strlen( new_wd ) + 1;
    netio_obj.Writen( sockfd, new_wd, len );
    cmd = netio_obj.RecvCommand( sockfd, prog_name );
    if (cmd == file_exists) {
      printf("Ok\n");
    }
    else if (cmd == no_such_file) {
      printf("no such path\n");
    }
    else {
      printf("%s: server_cd - unexpected command\n", prog_name );
    }
  }
} // server_cd



/*
 * print_server_cwd
 *
 
   Print the current working directory of the server.

 */
void ftp_client::print_server_cwd( int sockfd, const char *prog_name )
{
  char cwd_buf[ 128 ];
  COMMANDS cmd;

  cwd_buf[0] = '\0';

  // send the print_cwd command to the server
  netio_obj.SendCommand( sockfd, print_cwd );
  cmd = netio_obj.RecvCommand( sockfd, prog_name );
  if (cmd == file_exists) {
    if (netio_obj.read_block( sockfd, cwd_buf, sizeof( cwd_buf ) ) > 0) {
      printf("%s\n", cwd_buf );
    }
  }
  else if (cmd == no_such_file) {
    printf("error reading directory\n");
  }
  else {
    printf("%s: print_server_cwd - unexpected command\n", prog_name );
  }
} // print_server_cwd



/*
 * put_transfer
 *

   Move a file from the local system to the remote system.  That is
   srcname on the local system is moved to destname on the remote
   system.

   Inputs:
     sockfd   : socket file descriptor
     srcname  : name of the file on the local system
     destname : name of the file on the remote system

  A Note On the Open Mode

   On Windows systems lines of text are terminated by a CR NL (new
   line) pair.  POSIX lines are terminated by a CR.  The ISO standard
   defines a "binary" mode file file open.  This mode is largely
   unused on UNIX systems (Solaris and BSD).  But it is used by
   Microsoft, which will translate the data read, removing the CR
   characters.  This wrecks havoc when transferring binary files
   like tar files.

   The BTP programs supports two modes - ascii and binary.  In ascii
   mode, translation takes place.  In binary mode there is no
   translation.  This is controlled through the open mode in fopen
   where the "b" character specified binary mode.

 */ 
void ftp_client::put_transfer( int sockfd, const char *srcname, 
			       const char *destname, const char *prog_name )
{
  const char *OPEN_READ_ONLY = "r";
  const char *OPEN_READ_ONLY_BINARY = "rb";
  const char *mode;
  FILE *src_fp;

  assert( srcname != NULL );
  
  if (netio_obj.get_mode() == binary_mode) {
    mode = OPEN_READ_ONLY_BINARY;
  }
  else {
    mode = OPEN_READ_ONLY;
  }

  // open the source file on the local system
  if ((src_fp = iofunc.Fopen( srcname, mode, prog_name)) != NULL) {

    // open the destination file on the server
    if (open_server_dest(sockfd, destname, prog_name )) {
      file_obj.put_file( sockfd, src_fp, srcname, prog_name );
    }

    fclose( src_fp );
  }
} // put_transfer



/*
 * get_transfer
 *

   Move a file from the remote system onto the local system.  That
   is srcname on the remote system is moved to destname on the local
   system.

   Inputs:
     sockfd   : socket file descriptor
     srcname  : name of the source file on the remote system
     destname : name of the destination file on the local system

   ok = true
   if destname is a file on the local system
     ok = false
     if user gives permission to overwrite the file
        ok = true
   end if
   if ok 
     if srcfile can be opened on the server
        open dest file
        transfer source to dest
        close dest file
     end if
   end if
 
 */
void ftp_client::get_transfer( int sockfd, const char *srcname, 
			       const char *destname, const char *prog_name )
{
  Boolean OK;

  printf("get_transfer: srcname = %s, destname = %s\n", srcname, destname );

  // Check whether the destination file exists on the client side
  // and if so, whether its a directory or a file.
  OK = TRUE;
  if (iofunc.is_file( destname )) {
    // the file exists
    OK = FALSE;
    if (iofunc.is_dir( destname )) {
      printf("%s: %s is a directory\n", prog_name, destname );
    }
    else {
      // ask user if its OK to overwrite the file on the client
      if (ok_to_overwrite( destname )) {
	OK = TRUE;
      }
    }
  }

  // if the file does not exist on the client or its OK to overwrite
  // the file...
  if (OK) {
    if ( open_server_src( sockfd, srcname, prog_name ) ) {
      const char *WRITE = "w";
      const char *WRITE_BINARY = "wb";
      const char *mode;
      FILE *dest_fp;

      if (netio_obj.get_mode() == binary_mode) {
	mode = WRITE_BINARY;
      }
      else {
	mode = WRITE;
      }

      if ((dest_fp = iofunc.Fopen(destname, mode, prog_name)) != NULL) {
	time_interval interval;
	int bytes_transfered;

	interval.start_time();
	bytes_transfered = file_obj.get_file( sockfd, dest_fp, 
					      destname, prog_name );
	interval.end_time();
	fclose( dest_fp );
      }
    }
  }

} // get_transfer



static void *GetMem(const unsigned int size )
{
  return malloc( size );
}

static void FreeMem( void *pMem )
{
  free( pMem );
}


/*
 * get_server_names
 *

   The directory names sent from the server are sent in a
   io_block structure.  Each name is a null terminated string.

 */
void ftp_client::get_server_names( int sockfd, 
				   dirnames &dir, 
				   string_heap &strheap,
				   const char *prog_name )
{
  io_block blk;
  COMMANDS cmd;
  int byte_cnt, len;
  char *pStr, *pNewStr;

  while ((cmd = file_obj.get_io_block(sockfd, blk, prog_name)) == sending_data) {
    byte_cnt = 0;
    pStr = blk.buf;
    while (blk.size > byte_cnt) {
      len = strlen(pStr) + 1;
      pNewStr = strheap.alloc_string( pStr );
      dir.add_dir_name( pNewStr );
      byte_cnt = byte_cnt + len;
      pStr = pStr + len;
    }
  }
} // get_server_names



/*
 * server_ls
 *

   Display the files in the current directory on the server
   in UNIX ls format.

   Command:
      local ls         : list the files in the current working directory
      local ls <path>  : list the files in the <path> directory

 */
void ftp_client::server_ls( int sockfd, const char *prog_name )
{
  const int MAXPATH = 128;
  char path[ MAXPATH ];
  dirnames dir( GetMem, FreeMem );
  string_heap strheap;
  int len;

  netio_obj.SendCommand( sockfd, list_dir );
  netio_obj.GetAck( sockfd, list_dir, prog_name );

  // if no path was given, the path is "."
  if (! com.get_word( path, sizeof(path) ) ) {
    path[0] = '.';
    path[1] = '\0';
  }

  // send the path name to the server
  len = strlen( path ) + 1;
  netio_obj.Writen( sockfd, path, len );

  // get the file names from the server
  get_server_names( sockfd, dir, strheap, prog_name );
  if (dir.names.get_size() > 0) {
    dir.print_names();
  }
  else {
    printf("%s: no such file or directory\n", prog_name );
  }
} // server_ls



/*
 * client_ls
 *

   List a directory on the client.

   Command:
      local ls         : list the files in the current working directory
      local ls <path>  : list the files in the <path> directory

 */
void ftp_client::client_ls( const char *progname )
{
  const int MAXPATH = 128;
  char path[ MAXPATH ];

  // if no path was given, get the current working directory
  if (! com.get_word( path, sizeof(path) ) ) {
    net_services netserv;

    if (netserv.GetCwd(path, sizeof(path) ) == NULL) {
      printf("%s: get cwd failed\n", progname );
      return; // RETURN
    }
  }

  // At this point we have a directory path - either
  // a directory path entered by the user or the cwd.
  readdir dread;
  dirnames dir( GetMem, FreeMem );
  char *name;

  for (name = dread.first_file( path ); name != NULL; name = dread.next_file())
    dir.add_dir_name( name );

  dir.print_names();
} // client_ls



/*
 * local_cd
 *

   Change the current working directory on the client.

   Command:
       local cd <path>

 */
void ftp_client::local_cd( const char *prog_name )
{
  const int MAXPATH = 80;
  char new_dir[ MAXPATH ];

  if (com.get_word( new_dir, sizeof(new_dir) )) {
    if (chdir( new_dir) == 0) {
      printf("Ok\n");
    }
    else {
      printf("%s: cd failed for %s\n", prog_name, new_dir );
    }
  }
  else {
    printf("%s: path expected\n", prog_name );
  }
} // local_cd



/*
 * print_local_cwd
 *

   command:
      local pwd

   Execute the "pwd" (print working directory) command
   on the client side.
 
 */
void ftp_client::print_local_cwd(const char *prog_name )
{
  net_services netserv;
  char dirbuf[128];

  if (netserv.GetCwd( dirbuf, sizeof(dirbuf) ) != NULL) {
    printf("client: %s\n", dirbuf );
  }
  else {
    printf("%s: get cwd failed\n", prog_name );
  }
  
} // print_local_cwd;



/*
 * do_local_commands
 *

   Execute commands that only effect the client (no
   interaction from the server).

 */
void ftp_client::do_local_commands( const char *prog_name )
{
  const int MAXWRD = 80;
  char cmd[ MAXWRD ];
  COMMANDS local_cmd;

  if (com.get_word( cmd, sizeof(cmd) )) {
    local_cmd = com.find_command( cmd );

    switch (local_cmd) {
    case change_dir:
      local_cd( prog_name );
      break;

    case print_cwd:
      print_local_cwd( prog_name );
      break;

    case list_dir:
      client_ls( prog_name );
      break;

    default:
      printf("%s: unknown local command = %s\n", prog_name, cmd );
      break;
    } // switch
  }
  else {
    printf("%s: local command expected\n", prog_name );
  }

} // do_local_commands


/*
 * debug_command
 *

   Process the command line command to turn on or turn off
   debug command.

 */
void ftp_client::debug_command( int sockfd, const char *prog_name )
{
  const int MAXWRD = 80;
  char debug_arg[ MAXWRD ];
  COMMANDS cmd = debug_mode_on;
  
  if (com.get_word( debug_arg, sizeof(debug_arg))) {
    cmd = com.find_command( debug_arg );
    if (cmd == turn_on)
      cmd = debug_mode_on;
    else if (cmd == turn_off)
      cmd = debug_mode_off;
    else
      cmd = bad_command;
  }

  if (! bad_command) {
    if (cmd == debug_mode_on)
      file_obj.set_debug_mode();
    else
      file_obj.clear_debug_mode();

    netio_obj.SendCommand( sockfd, cmd );
    netio_obj.GetAck( sockfd, cmd, prog_name );
  }
} // debug_command



void ftp_client::title(void)
{
  printf("Bear Transfer Protocol (BTP) file transfer utility\n");
}


/*
 * copyright
 *
 */
void ftp_client::copyright()
{
  printf("Copyright: Ian Kaplan, all rights reserved, %d\n", year );
} // copyright



void ftp_client::version()
{
  const size_t major_version = 1;
  const size_t minor_version = 1;

  printf("BTP Version %d.%d, %s %d, %d\n", 
	 major_version, minor_version,
	 month, day, year );
}


/*
 * print_about
 *

   Print information about author and allowed use.
 
 */
void
ftp_client::print_about()
{
  title();
  printf("Author:    Ian Kaplan\n");
  copyright();
  printf("\n");
  printf("BTP is not public-domain software, shareware, and it is not protected\n");
  printf("by a 'copyleft' agreement, like the code from the Free Software\n");
  printf("Foundation.\n");
  printf("\n");
  printf("BTP is available free for your personal research and instructional use\n");
  printf("under the 'fair use' provisions of the copyright law. You may,\n");
  printf("however, redistribute BTP in whole or in part provided you acknowledge\n");
  printf("its source and include this copyright. You may, for example, include\n");
  printf("the distribution in a CDROM of free software, provided you charge only\n");
  printf("for the media, or mirror the distribution files at your site.\n");
  printf("\n");
  printf("You may not sell BTP or any product derived from it in which it is a\n");
  printf("significant part of the value of the product.\n");
  printf("\n");
  printf("For commercial use contact iank@bearcave.com (www.bearcave.com)\n");
  printf("\n");
  version();
  printf("\n");
} // print_about



void
ftp_client::help_summary()
{
  printf("get source_file [dest_file]\n");
  printf("put source_file [dest_file]\n");
  printf("pwd\n");
  printf("cd\n");
  printf("ls\n");
  printf("ascii\n");
  printf("binary\n");
  printf("about\n");
  printf("local command\n");
  printf("debug [on|off]\n");
  printf("help [command]\n");
  printf("quit\n");
} // help_summary


/*
 * print_help_display
 *

   Print a brief explaination for the supported commands.

 */
void
ftp_client::print_help_display(void)
{
  const int MAXWRD = 80;
  char help_arg[ MAXWRD ];

  if (com.get_word(help_arg, sizeof(help_arg))) {
    COMMANDS cmd = com.find_command( help_arg );

    switch (cmd) {
    case quit:
      printf("quit\n");
      printf("Send a message to the server to exit and exit the client\n");
      break;
    case get_file:
      printf("get file_name [dest_name]\n");
      printf("Get a file from the remote system.  If a destination\n");
      printf("file name is given, write to ""dest_name"", otherwise\n");
      printf("write to ""file_name""\n");
      break;
    case put_file:
      printf("put file_name [dest_name]\n");
      printf("Sent a file to the remote system.  If a destination name\n");
      printf("is given, the file will be written to ""dest_name"" on the\n");
      printf("remote system.  Otherwise the file will be written to\n");
      printf("the file ""file_name""\n");
      break;
    case print_cwd:
      printf("pwd\n");
      printf("Print the server's current working directory\n");
      break;
    case change_dir:
      printf("cd path\n");
      printf("Change the server's current working directory to ""path""\n");
      break;
    case list_dir:
      printf("ls [path]\n");
      printf("List the ""path"" file or directory.  If no path is given\n");
      printf("list the server's current working directory.  The ls command\n");
      printf("uses UNIX 'ls -F' format\n");
      break;
    case local_command:
      printf("local command\n");
      printf("Execute a directory command on the client side.  These\n");
      printf("commands are: pwd, cd and ls\n");
      break;
    case ascii_mode:
      printf("ascii\n");
      printf("Ascii mode.  In ascii mode there may be file translation\n");
      printf("when a file is transferred between UNIX and NT.  File \n");
      printf("translation converts the Windows CR LF (0x0d 0x0a) to \n");
      printf("CR (0x0d).  This mode should not be used for binary file, \n");
      printf("zip file or tar file transfer\n");
      break;
    case binary_mode:
      printf("binary\n");
      printf("Binary mode.  In binary mode there is no file translation\n");
      printf("when a file is transferred between systems.  The memory \n");
      printf("image of the file, as it is read from disk, is transferrred\n");
      printf("between systems.\n");
      break;
    case debug_mode:
      printf("debug [on|off]\n");
      printf("""debug"" or ""debug on"" turns debug mode on\n");
      printf("""debug off"" turns debug mode off\n");
      break;
    case help:
      printf("Help, I need somebody, help, not just any body...\n");
      break;
    case about:
      printf("Print author and copyright information\n");
      break;
    default:
      break;
    }
  }
  else {
    help_summary();
  }
} // print_help_display



/*
 * process_commands
 *

   ftp command interpreter

 */
void
ftp_client::process_commands( int sockfd, const char *prog_name )
{
  const int MAXWRD = 80;
  char srcfile[ MAXWRD ], destfile[ MAXWRD ];
  int done;
  COMMANDS cur_command;

  done = FALSE;

  title();
  copyright();// print BTP copyright and version
  version();   
  netio_obj.print_mode();
  printf(" mode\n");

  while (! done ) {
    com.get_line();
    cur_command = com.get_command();

    switch (cur_command) {

    // put|get <source_file> [<dest_file>]
    case put_file:
    case get_file:
      if (com.get_word(srcfile, sizeof(srcfile))) {
	if (com.get_word(destfile, sizeof(destfile))) {
	  if (cur_command == put_file)
	    put_transfer( sockfd, srcfile, destfile, prog_name );
	  else
	    get_transfer( sockfd, srcfile, destfile, prog_name );
	}
	else {
	  if (cur_command == get_file)
	    get_transfer( sockfd, srcfile, srcfile, prog_name );
	  else
	    put_transfer( sockfd, srcfile, srcfile, prog_name );
	}
      }
      else {
	printf("%s: source file name expected\n", prog_name );
      }
      break;

    case change_dir:
      server_cd( sockfd, prog_name );
      break;

    case print_cwd:
      print_server_cwd( sockfd, prog_name );
      break;

    case ascii_mode:
      netio_obj.set_mode( ascii_mode );
      netio_obj.SendCommand( sockfd, ascii_mode );
      netio_obj.print_mode();
      printf(" mode\n");
      break;

    case binary_mode:
      netio_obj.set_mode( binary_mode );
      netio_obj.SendCommand( sockfd, binary_mode );
      netio_obj.print_mode();
      printf(" mode\n");
      break;

    case list_dir:
      server_ls( sockfd, prog_name );
      break;

    // commands affecting the server
    case local_command:
      do_local_commands( prog_name );
      break;

    case debug_mode:
      debug_command( sockfd, prog_name );
      break;

    case help:
      print_help_display();
      break;

    case about:
      print_about();
      break;

    case quit:
      netio_obj.SendCommand( sockfd, quit );
      done = TRUE;
      break;

    default:
      printf("process_commands: unknown command\n");
      break;

    } // switch
  } // while
} // process_commands



