
/* Bear Transfer Protocol (BTP), a TCP/IP socket based file 
   transfer utility

  The author of this software is Ian Kaplan
  www.bearcave.com
  iank@bearcave.com

  Copyright (c) Ian Kaplan, 1999

  See "copyright" file for allowed use.

*/
/*

   Bear Transfer Protocol: The Bear Products International
                           file transfer program.

   BTP Server

 */

/*
   To compile use Makefile

 */

#include "netinc.h"

#include "neterr.h"
#include "netwrap.h"
#include "iowrap.h"
#include "commands.h"
#include "netio.h"
#include "interval.h"
#include "netfile.h"
#include "netserv.h"
#include "switch.h"
#include "list.h"
#include "strheap.h"
#include "readdir.h"
#include "comline.h"
#include "ftp_server.h"



/*
 * get_dest_name
 *

   Get the destination file name from the client.  If
   the transaction goes OK, return a pointer to the
   file_name argument, otherwise return NULL;

 */
const char *ftp_server::get_dest_name(int sockfd, char *file_name, 
				      const int name_size,
				      const char *prog_name )
{
  COMMANDS sub_com;  // sub-command
  char *NamePtr;

  NamePtr = NULL;

  // get send_name command
  sub_com = netio_obj.RecvCommand( sockfd, prog_name );

  if (sub_com == send_name) {
    // send acknowledge
    netio_obj.SendCommand( sockfd, acknowledge );

    // get file name
    if (netio_obj.read_block(sockfd, file_name, name_size ) > 0) {

      // Check to see if file exists.  If the file exists,
      // have the client ask the user if the file should
      // be overwritten
      if (iofunc.is_file( file_name )) {

	// if its not a directory
	if (! iofunc.is_dir( file_name)) {
	  netio_obj.SendCommand( sockfd, file_exists );

	  // wait for response to file_exists: overwrite_file or
	  // acknowledge
	  sub_com = netio_obj.RecvCommand( sockfd, prog_name );
	  if (sub_com == overwrite_file) {
	    NamePtr = file_name;
	    netio_obj.SendCommand( sockfd, acknowledge );
	  }
	}
	else {
	  // file is a directory
	  netio_obj.SendCommand( sockfd, directory );
	}
      }
      else {
	// file does not exist, so it OK to try opening it
	NamePtr = file_name;
	netio_obj.SendCommand( sockfd, acknowledge );
      }
    }
    else {
      // communication error
      netio_obj.SendCommand( sockfd, com_error );
    }
  }

  return NamePtr;
} // get_dest_name




/*
 * get_src_name
 *

   Get the source file name (e.g., the file on the server) from the
   client.  If the transaction goes OK, return a pointer to the
   file_name argument, otherwise return NULL;

 */
const char *ftp_server::get_src_name(int sockfd, char *file_name, 
				     const int name_size,
				     const char *prog_name )
{
  COMMANDS sub_com;  // sub-command
  char *NamePtr;

  NamePtr = NULL;

  // get send_name command
  sub_com = netio_obj.RecvCommand( sockfd, prog_name );

  if (sub_com == send_name) {
    // send acknowledge
    netio_obj.SendCommand( sockfd, acknowledge );

    // get file name
    if (netio_obj.read_block(sockfd, file_name, name_size ) > 0) {

      // Check to see if file exists.  If it exists, make sure that
      // its not a directory
      if (iofunc.is_file( file_name )) {

	if (iofunc.is_dir( file_name)) {
	  // file is a directory
	  netio_obj.SendCommand( sockfd, directory );
	}
	else {
	  // File is not a directory.
	  // Note that the acknowledge is not sent until
	  // the file is successfully opened.
	  NamePtr = file_name;
	}
      }
      else {
	// file does not exist
	netio_obj.SendCommand( sockfd, no_such_file );
      }
    }
    else {
      // communication error
      netio_obj.SendCommand( sockfd, com_error );
    }
  }

  return NamePtr;
} // get_src_name



/*
 * change_cwd
 *

   Get a directory path from the client and attempt to
   do a chdir to change the current working directory.

 */
void
ftp_server::change_cwd( int sockfd )
{
  char dirbuf[128];

  // send acknowledge
  netio_obj.SendCommand( sockfd, acknowledge );

  // get the directory path
  if (netio_obj.read_block( sockfd, dirbuf, sizeof(dirbuf) ) > 0) {
    // got path, now attempt the chdir
    if (chdir( dirbuf ) == 0) {
      netio_obj.SendCommand( sockfd, file_exists );
    }
    else {
      netio_obj.SendCommand( sockfd, no_such_file );
    }
  }
} // change_cwd




/*
 * send_cwd
 *

   Send the current working directory to the client.

 */
void
ftp_server::send_cwd( int sockfd )
{
  net_services netserv;
  char dirbuf[128];
  char *pChar;
  
  pChar = netserv.GetCwd( dirbuf, sizeof(dirbuf));
  if (pChar != NULL) {
    int len;

    netio_obj.SendCommand(sockfd, file_exists);
    len = strlen( dirbuf ) + 1;  // include the null terminator
    netio_obj.Writen( sockfd, dirbuf, len );
  }
  else {
    netio_obj.SendCommand(sockfd, no_such_file);
  }
} // send_cwd (current working directory)



/*
 * put_dir_name
 *

   Put a name into an io_block structure, which will be sent to the
   server.  While the "com" argument is "sending_data", names are
   added to the structure until it is full.  The structure is then
   sent to the server with the command "sending_data".  When the "com"
   argument is "send_done", the io_block structure is sent if there
   are unsent names.  Then the "send_done" command is sent.  When "com"
   is "send_done", name should be NULL.

 */
void 
ftp_server::put_dir_name( int sockfd, char *name, io_block &blk, COMMANDS com )
{
  if (com == sending_data) {
    assert( name != NULL );

    int num_bytes = strlen( name ) + 1;
    
    if (blk.size + num_bytes > sizeof(blk.buf)) {
      netio_obj.send_io_block( sockfd, blk, blk.size );
      blk.size = 0;
    }
    
    strcpy( &blk.buf[ blk.size ], name );
    blk.size = blk.size + num_bytes;
  }
  else if (com == send_done) {
    if (blk.size > 0) {
      netio_obj.send_io_block( sockfd, blk, blk.size );
    }
    netio_obj.SendCommand( sockfd, send_done );
  }
} // put_dir_name



/*
 * send_dir_names
 *
 */
void
ftp_server::send_dir_names( int sockfd )
{
  const int MAXPATH = 128;
  char path[ MAXPATH ];

  // send an acknowledge for the list_dir command
  netio_obj.SendCommand( sockfd, acknowledge );

  // Read the path.  If the local directory is to be listed
  // then the path will be "."
  if (netio_obj.read_block( sockfd, path, sizeof(path) ) > 0) {
    io_block blk;
    readdir dread;
    char *name;

    blk.size = 0;
    blk.com = sending_data;

    for (name = dread.first_file( path ); 
	 name != NULL; 
	 name = dread.next_file()) {
      put_dir_name( sockfd, name, blk, sending_data );
    }
    put_dir_name( sockfd, NULL, blk, send_done );
  }
  else {
    printf("send_dir_names: could not read path from server\n");
  }
} // send_dir_names




/*
 * process_commands
 *

   Process the commands sent via a socket from the server

 */
void
ftp_server::process_commands( int sockfd, const char *prog_name )
{
  const int MAX_NAME = 160;
  int read_rslt;
  netio nio;
  unsigned int done;
  COMMANDS com;         // user command
  char file_name[ MAX_NAME ];
 
  done = FALSE;
  while (! done ) {

    com = netio_obj.RecvCommand( sockfd, prog_name );

    switch (com) {

    // move a file from the server's system to the client system
    case get_file:
      // send acknowlege
      netio_obj.SendCommand( sockfd, acknowledge );
      if (get_src_name(sockfd, file_name, sizeof(file_name), 
		                                  prog_name) != NULL) {
	const char *OPEN_READ = "r";
	const char *OPEN_READ_BINARY = "rb";
	const char *mode;
	FILE *src_fp;

	if (netio_obj.get_mode() == binary_mode) {
	  mode = OPEN_READ_BINARY;
	}
	else {
	  mode = OPEN_READ;
	}

	if ((src_fp = fopen( file_name, mode )) != NULL) {
	  netio_obj.SendCommand( sockfd, acknowledge );
	  file_obj.put_file( sockfd, src_fp, file_name, prog_name );
	  fclose( src_fp );
	}
	else {
	  netio_obj.SendCommand( sockfd, could_not_open );
	}
	// try to open file
      }
      
      break;

    // move file from the client system to the server system
    case put_file:
      // send acknowlege
      netio_obj.SendCommand( sockfd, acknowledge );

      if (get_dest_name( sockfd, file_name, sizeof(file_name), 
			                           prog_name) != NULL) {
	const char *OPEN_WRITE_TRUNC = "w";
	const char *OPEN_WRITE_TRUNC_BINARY = "wb";
	const char *mode;
	FILE *dest_fp;

	if (netio_obj.get_mode() == binary_mode) {
	  mode = OPEN_WRITE_TRUNC_BINARY;
	}
	else {
	  mode = OPEN_WRITE_TRUNC;
	}

	// open file for write/truncate
	if ((dest_fp = fopen( file_name, mode )) != NULL) {
	  int bytes_read;

	  // OK to send source file data
	  printf("%s: socket input to %s\n", prog_name, file_name );

	  // Read socket input into file
	  bytes_read = file_obj.get_file( sockfd, dest_fp, 
					  file_name, prog_name );

	  // close file
	  fclose( dest_fp );
	}
	else {
	  // tell server that the destination file can't be opened
	  netio_obj.SendCommand( sockfd, could_not_open );
	}
      }
      break;

    case list_dir:
      send_dir_names( sockfd );
      break;

    case change_dir:
      change_cwd( sockfd );
      break;

    case print_cwd:
      send_cwd( sockfd );
      break;

    case ascii_mode:
      netio_obj.set_mode( ascii_mode );
      break;

    case binary_mode:
      netio_obj.set_mode( binary_mode );
      break;

    case debug_mode_on:
    case debug_mode_off:
      if (com == debug_mode_on) {
	file_obj.set_debug_mode();
      }
      else {
	file_obj.clear_debug_mode();
      }
      netio_obj.SendCommand( sockfd, acknowledge );
      break;

    case quit:
      done = TRUE;
      break;

    default:
      printf("process_commands: bad command = 0x%08x\n", com );
      break;

    } // switch

  } // while

} // process_commands



