
/* Bear Transfer Protocol (BTP), a TCP/IP socket based file 
   transfer utility

  The author of this software is Ian Kaplan
  www.bearcave.com
  iank@bearcave.com

  Copyright (c) Ian Kaplan, 1999

  See "copyright" file for allowed use.

*/

#include "netinc.h"
#include <sys/types.h>
#include <sys/stat.h>

#include "neterr.h"
#include "iowrap.h"



/*
 * Fopen
 *
 */
FILE *iowrap::Fopen( const char *file_name, const char *mode, 
		     const char *prog_name )
{
  FILE *fp;

  if ((fp = fopen( file_name, mode )) == NULL) {
    char buf[ 256 ];

    sprintf(buf, "%s: could not open file %s", prog_name, file_name );
    perror( buf );
  }

  return fp;
} // Fopen



/*
 * Fread
 *

   Read count bytes from the file pointed to by fp

 */
size_t iowrap::Fread( char *buf, size_t buf_size, FILE *fp,
		      const char *file_name, const char *prog_name )
{
  size_t bytes_read;

  bytes_read = fread( buf, sizeof(char), buf_size, fp );
  if (bytes_read < buf_size) {
    if (ferror( fp ) != 0) {
      char buf[ 256 ];

      sprintf(buf, "%s: error reading from file %s", prog_name, file_name );
      perror( buf );
    }
  }

  return bytes_read;
} // Fread



Boolean iowrap::Fwrite( char *buf, const int buf_size, FILE *fp, 
			const char *file_name, const char *prog_name )
{
  Boolean ok = TRUE;
  size_t write_rslt;

  write_rslt = fwrite( buf, sizeof(char), buf_size, fp );
  if (write_rslt != buf_size) {
    printf("%s: error writing to file %s\n", prog_name, file_name );
    ok = FALSE;
  }
  return ok;
} // Fwrite



void iowrap::Close( int fd )
{
  int close_rslt;

#ifdef WIN32
  close_rslt = closesocket( fd );
#else
  close_rslt = close( fd );
#endif

  if (close_rslt != 0)
    Error("Close error");
} // Close




char * iowrap::Fgets(char *ptr, int n, FILE *stream)
{
  char	*rptr = NULL;

  if ( (rptr = fgets(ptr, n, stream)) == NULL && ferror(stream))
    Error("fgets error");

  return (rptr);
} // Fgets



void iowrap::Fputs(const char *ptr, FILE *stream)
{
  if (fputs(ptr, stream) == EOF)
    Error("fputs error");
}  // Fputs



/*
 * is_file
 *

   The is_file function tests whether file_name is a file or
   directory.  The is_file function returns TRUE if the file exists,
   FALSE otherwise.

   This function uses the POSIX stat function and the struct stat data
   structure.

 */
Boolean iowrap::is_file( const char *file_name )
{
  struct stat stat_struct;
  Boolean rslt;
  int stat_rslt;
  
  assert( file_name != NULL );

  rslt = FALSE;

  stat_rslt = stat( file_name, &stat_struct);
  if (stat_rslt < 0) {
    rslt = FALSE;
    // the ENOENT errno indicates that the file does not exist
    if (errno != ENOENT) {
      perror("file_exists(): stat failed");
    }
  }
  else {
    // file exists
    rslt = TRUE;
  }

  return rslt;
}  // is_file




/*
 * is_dir
 *
  
   Function returns TRUE if file_name is a directory
   FALSE otherwise.

 */
Boolean iowrap::is_dir( const char *file_name )
{
  struct stat stat_struct;
  Boolean rslt;
  int stat_rslt;
  
  assert( file_name != NULL );

  rslt = FALSE;

  stat_rslt = stat( file_name, &stat_struct);
  if (stat_rslt < 0) {
    rslt = FALSE;
    // the ENOENT errno indicates that the file does not exist
    if (errno != ENOENT) {
      perror("file_exists(): stat failed");
    }
  }
  else {
    // file exists
    if (stat_struct.st_mode & S_IFDIR)
      rslt = TRUE;
  }

  return rslt;
} // is_dir
