
/* Bear Transfer Protocol (BTP), a TCP/IP socket based file 
   transfer utility

  The author of this software is Ian Kaplan
  www.bearcave.com
  iank@bearcave.com

  Copyright (c) Ian Kaplan, 1999

  See "copyright" file for allowed use.

*/


#include "netinc.h"
#include "neterr.h"
#include "commands.h"
#include "iowrap.h"
#include "netio.h"


/*
    A Note on "Endian-ness"

      Our histories of six thousand moons make no mention of any other
      regions, than the two great empires of Lilliput and Blefuscu.
      These two mighty powers have, as I was going to tell you, been
      engaged in a most obstinate war for six and thirty moons past.
      It began upon the following occasion.

      It is allowed on all hands, that the primative way of breaking
      eggs before we eat them, was upon the larger end: but his
      present Majesty's grandfather, while he was a boy, going to eat
      an egg, and breaking it according to the ancient practice,
      happened to cut one of his fingers.  Whereupon the emperor his
      father, published an edict, commanding his subjects, upon great
      penalty, to break the smaller end of their eggs.

      [...]

      It is computed, that eleven thousand persons have, at several
      times, suffered death, rather than submit to break their eggs at
      the smaller end.

      Gulliver's Travels, Jonathan Swift

    On a "little endian" machine a pointer to a word of memory points
    to the least significant byte.

    On a "big endian" machine a pointer to a word of memory points to
    the most significant byte.

    The iNTEL x86 family is little endian.  The SPARC and powerPC are
    big endian.

    The server and the client may be running on systems with different
    endian-ness.  This matters when sending word size values,
    including COMMANDS and data sizes.  If the platforms the client
    and server are running on differ in endian-ness, the word values
    are converted when they are read.

 */

/*
 * client_endian_check
 *

   Pass the 32-bit value 0x00000001 to the server.  The server will
   pass back its version of 0x00000001.  If the values don't match
   then there is an endian mismatch between the client and the
   server.  As a result, the bytes must be flipped (see byte_flip,
   below) in each word passed between the systems.

   Note that this function assumes that both the client and the server
   are running on 32-bit machines or at least that the software was
   compiled for 32-bit mode.

 */
void netio::client_endian_check(int fd)
{
  // buffer for data from server
  size_t data_in[ DATA_BUF_SIZE/sizeof(size_t) ];
  int amt_read;
  size_t word_out = 0x00000001;

  Writen( fd, &word_out, sizeof( word_out ) );
  
  data_in[0] = 0;

  if ((amt_read = read_block( fd, (char *)data_in, sizeof( data_in ))) == 0) {
    Error("client_endian_check: server terminated prematurely");
  }
  else {
    if (amt_read == sizeof( word_out )) {
      if (data_in[0] != 0) {
	if (word_out != data_in[0]) {
	  // There is an endian mismatch between the client
	  // and the server.
	  mismatch = TRUE;
	}
      }
      else {
	Error("client_endian_check: no data returned from server");
      }
    }
  }
} // client_endian_check



/*
 * server_endian_check
 *

   Send 0x00000001 to the client so that the client can
   determine its endianness.

 */
void netio::server_endian_check(int fd )
{
  // buffer for data from client
  size_t data_in[ DATA_BUF_SIZE/sizeof(size_t) ];
  int amt_read;

  data_in[0] = 0;

  if ((amt_read = read_block( fd, (char *)data_in, sizeof( data_in ))) == 0) {
    Error("server_endian_check: error on read");
  }
  else {
    if (amt_read == sizeof( size_t )) {
      if (data_in[0] != 0) {
	if (data_in[0] != 0x00000001 ) {
	  // There is an endian mismatch between the client
	  // and the server.
	  mismatch = TRUE;
	}
      }
      else {
	Error("server_endian_check: no data sent by client");
      }
    }  
  }

  // Send the client 0x00000001
  size_t word_out = 0x00000001;

  Writen( fd, &word_out, sizeof( word_out ) );
}  // server_endian_check




/*
 * byte_flip
 *

   Convert from one "endian-ness" to another "endian-ness"

   This function is passed an unsigned word in one format
   (little endian) and it returns it in another format
   (big endian).

 */
unsigned int netio::byte_flip( const unsigned int little )
{
  typedef union {
    unsigned char bytes[ 4 ];
    unsigned int u;
  } TRIX;

  TRIX big;
  unsigned char t;

  big.u = little;
  t = big.bytes[3];
  big.bytes[3] = big.bytes[0];
  big.bytes[0] = t;
  t = big.bytes[1];
  big.bytes[1] = big.bytes[2];
  big.bytes[2] = t;

  return big.u;
} // byte_flip



/*
 * writen
 *
   Write "n" bytes from the buffer pointed to by
   vptr to the file descriptor fd.

 */
int
netio::writen(int fd, const void *vptr, size_t n)
{
  const int FLAGS = 0;
  size_t     nleft;
  int        nwritten;
  const char *ptr;
  int        err_rslt;

  ptr = (char *)vptr;
  nleft = n;
  while (nleft > 0) {
    if ( (nwritten = send(fd, ptr, nleft, FLAGS )) <= 0) {
      err_rslt = get_errno();
      if (is_EINTR( err_rslt )) {
	/* send was canceled by an external signal, try again */
	printf("writen: is_EINTR is TRUE - send was canceled\n");
	nwritten = 0;
      }
      else
	return(-1);  /* error */
    }

    nleft -= nwritten;
    ptr   += nwritten;
  }
  return(n);
}
/* end writen */



/*
 * get_block
 *
  
   Read a block of data from the socket fd.

 */
int
netio::read_block(int fd, char *pBlock, unsigned int block_size )
{
  const int FLAGS = 0;
  int cnt = 0;

  do {
    cnt = recv(fd, pBlock, block_size, FLAGS);
  } while (cnt < 0 && is_EINTR( get_errno()) );
  
  return( cnt );
} /* read_block */




/*
 * Writen
 *

   Write N bytes, pointed to by ptr, to the socket descriptor "fd"

 */
void
netio::Writen(int fd, void *ptr, size_t nbytes)
{
  if (writen(fd, ptr, nbytes) != nbytes)
    Error("writen error");
} // writen



/*
 * send_io_block
 *

   Send an io_block data structure.

 */
void netio::send_io_block( int sockfd, io_block &blk, 
			   int num_bytes)
{

  if (debug) {
    printf("send_io_block: data size = %d, total sent = %d\n", 
	   num_bytes, num_bytes + IO_BUF_EXTRA );
  }

  // Note that the command and size are not sent byte flipped if there
  // is an endian mismatch.  The protocol is that the reader 
  // always puts the data in the right form.
  blk.com = sending_data;
  blk.size = num_bytes;

  // send data block
  Writen(sockfd, (void *)&blk, num_bytes + IO_BUF_EXTRA);
}  // send_io_block



/*
 * RecvCommand
 *

   Get a command response from the remote system

 */
COMMANDS netio::RecvCommand( int sockfd, const char *prog_name )
{
  char message_in[ DATA_BUF_SIZE ];
  neterr err;
  COMMANDS com;
  char *pWordBytes;
  int num_read, i;

  if ((num_read = read_block( sockfd, message_in, sizeof( message_in))) == 0) {
    char buf[128];
    
    sprintf(buf, "%s: RecvCommand - server terminated prematurely", 
	         prog_name );
    Error( buf );
  }

  if (num_read != sizeof( com ))
    printf("RecvCommand: num_read = %d\n", num_read );

  pWordBytes = (char *)&com;

  // If there is no mismatch between the endian-ness, the
  // data can simply be copied across
  if (! endian_mismatch()) {
    for (i = 0; i < sizeof(com); i++) {
      pWordBytes[i] = message_in[ i ];
    }
  }
  else { // There is an endian mismatch, so flip bytes
    for (i = 0; i < sizeof(com); i++) {
      pWordBytes[(sizeof(com)-1) - i] = message_in[ i ];
    }
  }

  return com;
} // RecvCommand



/*
 * GetAck
 *
 */
void netio::GetAck( int sockfd, COMMANDS com, const char *prog_name )
{
  const int MAXBUF = 256;
  char buf[ MAXBUF ];
  client_commands comfx;
  COMMANDS com_in;
  neterr err;

  if ((com_in = RecvCommand( sockfd, prog_name )) != acknowledge) {
    sprintf(buf, "%s: GetAck failed for %s command", prog_name,
	    comfx.command_name( com ));
    err.Error( buf );    
  }
} // GetAck



/*
 * SendCommand
 *

   Send a command to the remote system

 */
void netio::SendCommand(int sockfd, COMMANDS com )
{
  Writen( sockfd, (char *)&com, sizeof( com ) );
} // SendCommand



/*
 * print_mode
 *

   Print the transfer mode (e.g., ascii or binary)

 */
void netio::print_mode(void)
{
  COMMANDS cmd;

  cmd = get_mode();
  if (cmd == ascii_mode)
    printf("ascii");
  else
    printf("binary");
} // print_mode



/* Initialize static variables that are shared
   by all instances of the netio class.
   */
Boolean  netio::mismatch = FALSE;
COMMANDS netio::mode = ascii_mode;
